//
//  Copyright (C) by S.Tomabechi 1999
//  All rights reserved.
//
//  Gaussian elimination
//


// return value : rank of matrix Z1
// nBit1    : #columns of Z1
// nBit2    : #columns of Z2
// nRow     : #rows
// eliminate the matrix (Z1,Z2)
#if 1

#ifdef __BORLANDC__
#pragma inline
#endif

unsigned long Gauss(unsigned long nBit1,unsigned long nBit2,
                                        unsigned long nRow,
                                        unsigned long** Z1,unsigned long** Z2)
{
    unsigned long j,k ;
    unsigned long nRank ;
    unsigned long nArray1,nArray2 ;

//  for( j=0 ; j<nBit1 ; j++ ) {
    // initialize
    __asm {
        mov ebx,Z1          // fix ebx as pointer to Z1

        xor eax,eax
        mov j,eax           // j <- 0
        mov nRank,eax       // nRank <- 0

        mov ecx,nBit1
        add ecx,31
        shr ecx,5
        mov nArray1,ecx                 // size of array

        mov ecx,nBit2
        or  ecx,ecx
        jz  SET_NULL
        add ecx,31
        shr ecx,5
        mov nArray2,ecx
        jmp OUTER_LOOP
    }
SET_NULL:
    __asm {
        mov Z2,ecx
    }

OUTER_LOOP:
    __asm {
        mov eax,j
        cmp eax,nBit1                // j<nBit1 ?
        jae RET1                                 // terminate if j>=nBit1
    }

//      for( k=nRank ; k<nRow ; k++ ) {
//          if( BitIndex(Z1[k],j) ) break ;
//      }

    __asm mov ecx,nRank                      // ecx <- nRank

SEARCH_LOOP:
        __asm {
            cmp ecx,nRow                            // ecx<nrow ?
            jae NEXT_ROW
            mov edx,eax                          // edx <- j
            mov edi,eax
            shr edx,5                        //
            and edi,0000001fh                // mod 32
        }
LOOP1:
        __asm {
            mov esi,[ebx+4*ecx]         // Z1[ecx]
            bt  [esi+4*edx],edi         // Bit(Z1[ecx][edx],edi) ?
            jc  NON_ZERO
            inc ecx
            jmp SEARCH_LOOP
        }

NON_ZERO:
        __asm {                                     // nonzero entry is found
            mov k,ecx

            cmp ecx,nRow                // k<nRow
            jae NEXT_ROW
        }

LOOP2:
        __asm {
            inc ecx                                 // k++
            cmp ecx,nRow                // k<nRow ?
            jae SWAP

            mov eax,j
            mov edx,eax
            shr edx,5                                // j/32
            and eax,0000001fh                        // j mod 32

            mov esi,[ebx+4*ecx]  // esi=Z1[l]
            bt  [esi+4*edx],eax  // Bit(Z1[l][edx],eax)
            jnc LOOP2

            mov edi,k
            mov edi,[ebx+4*edi]  // edi=Z1[k]

            mov eax,nArray1
            dec eax
        }
LOOP21:
        __asm {
            mov edx,[edi+4*eax]
            xor [esi+4*eax],edx     // Z1[l][*]^=Z1[k][*]
            dec eax
            jns LOOP21
        }
LOOP21_END:
        __asm {
            mov eax,Z2
            or  eax,eax                                             // Null pointer ?
            jz  LOOP2

            mov esi,[eax+4*ecx]  // esi=Z2[l]

            mov edi,k
            mov edi,[eax+4*edi]  // edi=Z1[k]

            mov eax,nArray2
            dec eax
        }
LOOP22:
        __asm {
            mov edx,[edi+4*eax]
            xor [esi+4*eax],edx  // Z2[l][*]^=Z2[k][*]
            dec eax
            jns LOOP22
        }
IS_ZERO:
        __asm {
            jmp LOOP2
        }
SWAP:                                             // swap rows
        __asm {
            mov esi,k
            cmp esi,nRank                                                    // k!=nRank ?
            jz  DO_NOT_SWAP
        }

            __asm {
                mov  edi,nRank

                mov  eax,[ebx+4*esi]             // swap Z1[esi] and Z1[edi]
                xchg eax,[ebx+4*edi]
                mov  [ebx+4*esi],eax
            }

            __asm {
                mov eax,Z2
                or  eax,eax
                jz  DO_NOT_SWAP                     // pointer is null

                mov  edx,[eax+4*esi]        // swap Z2[esi] and Z2[edi]
                xchg edx,[eax+4*edi]
                mov  [eax+4*esi],edx

            }
DO_NOT_SWAP:
        __asm inc nRank
NEXT_ROW:
        __asm {
            inc j
            jmp OUTER_LOOP
        }

RET1:
    return nRank ;
}


#else

#include "TBit.hpp"
// original C++ code
unsigned long Gauss(unsigned long nBit1,unsigned long nBit2,unsigned long nRow,
                                        unsigned long** Z1,unsigned long** Z2)
{
    long j,k ;
    unsigned long l ;
    unsigned long* w ;
    TBit B1,B2 ;
    unsigned long nRank(0) ;

    for( j=0 ; j<nBit1 ; j++ ) {
        for( k=nRank ; k<nRow ; k++ ) {
            if( BitIndex(Z1[k],j) ) break ;
        }
        if( k>=nRow ) continue ;
        for( l=k+1 ; l<nRow ; l++ ) {
            if( BitIndex(Z1[l],j) ) {
                B1.SetBitString(nBit1,Z1[k]) ;
                B2.SetBitString(nBit1,Z1[l]) ;
                B2^=B1 ;
                if( Z2!=NULL ) {
                    B1.SetBitString(nBit2,Z2[k]) ;
                    B2.SetBitString(nBit2,Z2[l]) ;
                    B2^=B1 ;
                }
            }
        }
        if( k!=nRank ) {
            w=Z1[k] ;
            Z1[k]=Z1[nRank] ;
            Z1[nRank]=w ;
            if( Z2!=NULL ) {
                w=Z2[k] ;
                Z2[k]=Z2[nRank] ;
                Z2[nRank]=w ;
            }
        }
        nRank++ ;
    }

    return nRank ;
}
#endif
