//
//  Copyright (C) by S.Tomabechi 1997,2001
//  All rights reserved.
//
//  compute square by Karatsuba's method
//
//  faster ( a little ) than multiplication as a*a 
//

#ifdef __BORLANDC__
#pragma inline
#endif

#ifndef __cplusplus
#error Must use C++ for the type Mint
#endif

#include "Mint.hpp"
#include "FFT.hpp"
#include "Error.hpp"
#include "MemMan.hpp"
static long k,l ;

Mint& Mint::isSquare(const Mint& a)
{

#ifdef _MEMMAN_DEBUG
a.check(__FILE__,__LINE__) ;
#endif

  int anDigit(a.nDigit) ;

  if( a.nDigit==0 ) {       // a==0 ?
    nDigit=0 ;
    nDim&=_POSITIVE_ ;      // signature <- 1
    return *this ;
  }


  //
  // a=a2*t+a1        t=(2^32)^n
  // a^2=a2^2*t^2+((a2+a1)^2-a2^2-a1^2)*t+a1^2
  //

  if( anDigit>Square_digit ) {
//  if( anDigit>FFT_digit ) return isMul(a,a) ; // give a job to FFT
    long n ;
             // multiplication by Fast Fourier Transformation
//FFT_digit=1<<20 ;
    if( anDigit>FFT_digit ) {
        n=anDigit ;
        k=1 ;
        l=0 ;
        while( n>k ) { k<<=1 ; l++ ; }     // index such that 2^k>=n
                                 // 傫Ƃ͔ɕ

//        if( (k-n)>(n-k/2) ) {
        if( 4*k>5*n ) {
//cout << " k " << hex << k << " " << n << dec ;
          n=k>>1 ;
          n-- ;
          goto KaraTsuba ;
        }
        if( l<NUM_OF_FFT_SCHEME+_FFT_BIT_ ) {
          _fft[l-_FFT_BIT_].fft_mul(a,a,*this) ;
          return *this ;
        }
//      else return isToom3(a,b) ;
    }
//  if( bnDigit>=1280 ) return Toom(a,b) ;

    k=n=anDigit ;
    n>>=1 ;

KaraTsuba:


//  int n(anDigit+1) ;
    Mint s2(k,0) ;
    Mint s1(k,0) ;                // a=s1*t+b    t:radix
    Mint ab(k,0) ;                // ab:= a+b
//  n>>=1 ;
    __asm {                  // check exact length of a1
      mov  esi,a
      mov  ebx,n
#ifdef __BORLANDC__
      mov  edi,[esi]
#else
      mov  edi,[esi+DIGIT]
#endif
     }
SearchZero1:
    __asm {
      cmp  dword ptr [ebx*NBYTE+edi-NBYTE],0
      jne  FindNonZero1
      dec  ebx
      jnz  SearchZero1
    }
FindNonZero1:
    __asm {
#ifdef __BORLANDC__
      mov  [esi.nDigit],ebx
#else
      mov  [esi+NDIGIT],ebx
#endif
      mov   edi,dword ptr s2.digit
#ifdef __BORLANDC__
      mov   bl,[esi.nDim+3]
      mov   byte ptr s2.nDim[3],bl
      mov   esi,[esi]
#else
      mov   bl,byte ptr [esi+SIGN]
      mov   byte ptr [s2+SIGN],bl
      mov   esi,[esi+DIGIT]
#endif
      mov   ebx,n
      lea   esi,[esi+ebx*NBYTE]
      mov   ecx,anDigit
      sub   ecx,ebx
    }
//NonZero1:
    __asm {
      mov   dword ptr s2.nDigit,ecx
      dec   ecx
    }
Shift1:
    __asm {
      mov   eax,[ecx*NBYTE+esi]
      mov   [ecx*NBYTE+edi],eax
      dec   ecx
      jns   Shift1
    }

    ab.isAdd(s2,a) ;           // ab=a+b
    s2.isSquare(s2) ;          // s2=a^2
    s1.isSquare(a) ;           // s1=b^2
    __asm {
      mov ebx,a
      mov ecx,anDigit
#ifdef __BORLANDC__
      mov [ebx.nDigit],ecx
#else
      mov [ebx+NDIGIT],ecx
#endif
    }

    isSquare(ab) ;         // ab=a+b
    _vMtemp.isAdd(s1,s2) ;   // a^2+b^2
    ab.isSub(*this,_vMtemp) ;  //   2ab=(a+b)^2-a^2-b^2 
//  (ab.isSub(*this,s2)).isSub(ab,s1) ; // 2ab=(a+b)^2-a^2-b^2 
    isShift32(s2,n<<1) ;           // a^2*t^2
//    Swap(*this,s2[2*n]) ;      // *this=s2[2*n]
    _vMtemp.isShiftAdd(*this,ab,n) ;   // a^2*t^2+2*ab*t
//    isAdd(*this,ab[n]) ;
#ifdef _MEMMAN_DEBUG
check(__FILE__,__LINE__) ;
#endif

    return isAdd(_vMtemp,s1) ;       // a^2*t^2+2*ab*t+b^2 
  }
#ifdef _MEMMAN_DEBUG
check(__FILE__,__LINE__) ;
#endif
  int abnDigit=anDigit<<1 ; // ς̍ő包
  int Count ;
  unsigned long* pa ;
  unsigned long Carry ;
  unsigned long Carry2 ;
  unsigned long Carry3(0) ;
  unsigned long save_eax ;

#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
#endif

  if( this==&a ) {         // copy a
    _uMtemp=a ;
    pa=_uMtemp.digit ;
  }
  else pa=a.digit ;

#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
#endif



  __asm {
    mov esi,this

#ifdef __BORLANDC__
    mov eax,[esi.nDim]
#else
    mov eax,[esi+NDIM]
#endif
    and eax,_DIGITS_
    cmp eax,abnDigit
    ja  SKIP_ALLOC
  }

//  if( ndim()<=abnDigit )
  allocate(abnDigit) ;  // enlarge digits

SKIP_ALLOC:
#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
#endif

  __asm {
//  mov   ebx,pb
    mov   esi,pa

    mov   edi,this

    mov   ecx,anDigit

    mov   edi,[edi]     // this->digit

    inc   ecx
//mov ecx,abnDigit
    xor   eax,eax
  }
clear:                  // for( j=0 ; j<ndigit ; j++ ) digit[j]=0 ;
  __asm {
    mov   [ecx*NBYTE+edi-NBYTE],eax   // 
    dec   ecx
    jnz   clear

  }
OuterLoop:
  __asm {
    mov   ecx,[esi]     // ecx=a.digit[j]     multiplier
  }
                        // case b.digit[j]!=0
InnerLoop:              // compute b.digit[j]*this
  __asm {
    mov   edx,anDigit
    xor   ebx,ebx
    mov   Count,edx
//  mov   Carry,ebx
    mov   Carry2,ebx
//  mov   Carry3,ebx


    mov   eax,ecx
    mul   ecx

    add   [ebx*NBYTE+edi],eax
    adc   edx,0     // オ

    inc   ebx
    mov   Carry,edx // save oferflow
    dec   Count
#if 0   
    jnz   Multiply
    add   [ebx*NBYTE+edi],edx
//  jmp   end_InnerLoop
    jmp   CheckDigits
#else       
    jz    CheckDigits 
#endif      

  }
Multiply:
  __asm {
      mov   eax,[ebx*NBYTE+esi]
      mul   ecx

//    push  eax
      mov   save_eax,eax
//    push  edx
#if 1
      xor   eax,eax               // compute [edx:eax] *=2 
      shld  eax,edx,1
      mov   Carry3,eax
#endif
//    pop   eax
      mov eax,save_eax
      shld  edx,eax,1
//    setc  byte ptr Carry3
      shl   eax,1

      add   eax,Carry   //  propagate carry
      adc   edx,0       //  add carry
                        //  overflow is not occured here
      add   [ebx*NBYTE+edi],eax

      adc   edx,0       // add carry
      mov   eax,Carry3
      add   edx,Carry2
      adc   eax,0
//    adc   Carry3,0

//      mov   eax,Carry3                 
      mov   Carry,edx
      mov   Carry2,eax


      inc   ebx
      dec   Count
      jnz   Multiply
//    mov   edx,Carry
      add   [ebx*NBYTE+edi],edx
      adc   eax,0
      inc   ebx

      mov   [ebx*NBYTE+edi],eax

   }
end_InnerLoop:
  __asm {
    lea     edi,[edi+2*NBYTE]
    lea     esi,[esi+NBYTE]
    dec     anDigit
    jnz     OuterLoop
   }

CheckDigits:
  __asm {
#if 1   
    add   [ebx*NBYTE+edi],edx
#endif  

    mov   edi,this
    mov   al,1
#ifdef __BORLANDC__
    mov   esi,[edi]
    mov   [edi.nDim+3],al       // result is positive
#else
    mov   esi,[edi]
    mov   [edi+SIGN],al
#endif
    mov   ebx,abnDigit
  }
IsZero:                       // search nonzero digit
  __asm {
    cmp   dword ptr [ebx*NBYTE+esi-NBYTE],0
    jne   Set
    dec   ebx
  }
Set:
OK:
  __asm {
#ifdef __BORLANDC__
    mov [edi.nDigit],ebx
#else
    mov [edi+NDIGIT],ebx
#endif
  }

RET1:
#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
check(__FILE__,__LINE__) ;
#endif
  return *this ;


}



#include <stdlib.h>
#include <iostream.h>
#include "CPUTime.hpp"
#include "Error.hpp"





#if 0
Mint& Mint::isSquare2(const Mint& a)
{

#ifdef _MEMMAN_DEBUG
a.check(__FILE__,__LINE__) ;
#endif

  int anDigit(a.nDigit) ;

  if( a.nDigit==0 ) {       // a==0 ?
    nDigit=0 ;
    nDim&=_POSITIVE_ ;      // signature <- 1
    return *this ;
  }


  //
  // a=a2*t+a1        t=(2^32)^n
  // a^2=a2^2*t^2+((a2+a1)^2-a2^2-a1^2)*t+a1^2
  //

  if( anDigit>Square_digit ) {
    if( anDigit>FFT_digit ) return isMul(a,a) ; // give a job to FFT
    int n(anDigit+1) ;
    Mint s2(n,0) ;
    Mint s1(n,0) ;                // a=s2*b+s1   b:radix
    Mint ab(n,0) ;                // ab:= a+b
    n>>=1 ;
    __asm {                  // check exact length of a1
      mov  esi,a
      mov  ebx,n
#ifdef __BORLANDC__
      mov  edi,[esi]
#else
      mov  edi,[esi+DIGIT]
#endif
     }
SearchZero1:
    __asm {
      cmp  dword ptr [ebx*NBYTE+edi-NBYTE],0
      jne  FindNonZero1
      dec  ebx
      jnz  SearchZero1
    }
FindNonZero1:
    __asm {
#ifdef __BORLANDC__
      mov  [esi.nDigit],ebx
#else
      mov  [esi+NDIGIT],ebx
#endif
      mov   edi,dword ptr s2.digit
#ifdef __BORLANDC__
      mov   bl,[esi.nDim+3]
      mov   byte ptr s2.nDim[3],bl
      mov   esi,[esi]
#else
      mov   bl,byte ptr [esi+SIGN]
      mov   byte ptr [s2+SIGN],bl
      mov   esi,[esi+DIGIT]
#endif
      mov   ebx,n
      lea   esi,[esi+ebx*NBYTE]
      mov   ecx,anDigit
      sub   ecx,ebx
    }
//NonZero1:
    __asm {
      mov   dword ptr s2.nDigit,ecx
      dec   ecx
    }
Shift1:
    __asm {
      mov   eax,[ecx*NBYTE+esi]
      mov   [ecx*NBYTE+edi],eax
      dec   ecx
      jns   Shift1
    }

    ab.isAdd(s2,a) ;           // ab=a2+a1
    s2.isSquare(s2) ;          // s2=a2^2
    s1.isSquare(a) ;           // s1=a1^2
    __asm {
      mov ebx,a
      mov ecx,anDigit
#ifdef __BORLANDC__
      mov [ebx.nDigit],ecx
#else
      mov [ebx+NDIGIT],ecx
#endif
    }

    ab.isSquare(ab) ;          // ab=a2+a1
    (ab.isSub(ab,s2)).isSub(ab,s1) ;
    isShift32(s2,n<<1) ;
//    Swap(*this,s2[2*n]) ;      // *this=s2[2*n]
    isShiftAdd(*this,ab,n) ;
//    isAdd(*this,ab[n]) ;
#ifdef _MEMMAN_DEBUG
check(__FILE__,__LINE__) ;
#endif

    return isAdd(*this,s1) ;
  }
  else {
#ifdef _MEMMAN_DEBUG
check(__FILE__,__LINE__) ;
#endif
  int abnDigit=anDigit<<1 ; // ς̍ő包
  int bnDigit=anDigit ;
  int Count ;
  unsigned long* bi ;
  unsigned long* pa ;
  unsigned long* pb ;
  unsigned long Carry ;
  unsigned long Carry2 ;
  unsigned long Carry3 ;
  signed char this_sign ;

#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
#endif

  if( this==&a ) {         // copy a
    _uMtemp=a ;
    pb=pa=_uMtemp.digit ;
  }
  else pb=pa=a.digit ;

#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
#endif


#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
#endif

  __asm {
    mov esi,a
#ifdef __BORLANDC__
    mov al,[esi.nDim+3]
    mov esi,a
    imul byte ptr [esi.nDim+3]
#else
    mov al,byte ptr [esi+SIGN]
    mov esi,b
    imul byte ptr [esi+SIGN]
#endif
    mov this_sign,al
    mov esi,this

#ifdef __BORLANDC__
    mov eax,[esi.nDim]
#else
    mov eax,[esi+NDIM]
#endif
    and eax,_DIGITS_
    cmp eax,abnDigit
    ja  SKIP_ALLOC
  }

//  if( ndim()<=abnDigit )
  allocate(abnDigit) ;  // g

SKIP_ALLOC:
#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
#endif

  __asm {
    mov   ebx,pb
    mov   esi,pa
    mov   edi,this
#ifdef __BORLANDC__
    mov   edi,[edi]
#else
    mov   edi,[edi]
#endif

    mov   bi,ebx        // save index

    mov   ecx,anDigit
mov ecx,abnDigit
    xor   eax,eax
  }
clear:                  // for( j=0 ; j<ndigit ; j++ ) digit[j]=0 ;
  __asm {
    mov   [ecx*NBYTE+edi-NBYTE],eax
    dec   ecx
    jnz   clear

  }
OuterLoop:
  __asm {
    mov   ecx,[esi]     // ecx=b.digit[j]
  }
                        // case b.digit[j]!=0
InnerLoop:              // compute b.digit[j]*this
  __asm {
    mov   edx,bnDigit
    mov   Count,edx
    xor   ebx,ebx

    mov   eax,ecx
    mul   ecx


    add   eax,Carry
    adc   edx,0

    add   [ebx*NBYTE+edi],eax
    adc   edx,0     // オ
 //   adc   Carry2,0
    mov   Carry,edx // save oferflow
//    adc   Carry2,0

    inc  ebx
    dec  Count
    jnz  Multiply
//    add  edx,Carry
//  xor eax,eax
    add  [ebx*NBYTE+edi],edx
    adc  Carry2,0
    mov  eax,Carry2
    inc ebx
    mov  [ebx*NBYTE+edi],eax
#if 0
    jnc  end_InnerLoop
    mov eax,1
    inv ebx
    mov  [ebx*NBYTE+edi],eax
#endif
//    adc   eax,0
//    add   Carry,eax
//    adc   Caey2,0
//    adc   Caey2,0
//    add   Carry2,edx
    jmp  end_InnerLoop

  }
Multiply:
  __asm {
      mov   eax,[ebx*NBYTE+esi]
      mul   ecx

      push  eax
      push  edx

      add   eax,Carry   //  propagate carry
      adc   edx,0       //  add carry
                        //  overflow is not occured here
      add   [ebx*NBYTE+edi],eax

      adc   edx,0       // オ
      add   Carry2,edx  // save oferflow

      adc   Carry3,0


      pop   edx
      pop   eax

//    add   eax,Carry   //  propagate carry
//    adc   edx,0       //  add carry
                        //  overflow is not occured here
      add   [ebx*NBYTE+edi],eax
      adc   edx,0       // オ

      add   Carry2,edx  // save oferflow
      adc   Carry3,0

      mov   eax,Carry2
      mov   Carry,eax

      mov   eax,Carry3
      mov   Carry2,eax

      mov Carry3,0

//  mov Carry2,0
//      adc   Carry,edx
      inc   ebx
      dec   Count
      jnz   Multiply
      xor   edx,edx
      mov   eax,Carry
      add   [ebx*NBYTE+edi],eax
      adc   Carry2,0
      add   edx,Carry2
      inc ebx

      mov   [ebx*NBYTE+edi],edx

//    mov   edx,Carry2
//    mov   Carry,edx
      mov   Carry,0
      mov   Carry2,0



   }
end_InnerLoop:
  __asm {
//  add     bi,NBYTE
//  add     bi,NBYTE
    lea     edi,[edi+2*NBYTE]
    lea     esi,[esi+NBYTE]
    dec     bnDigit
    jnz     OuterLoop
   }

CheckDigits:
  __asm {

    mov   edi,this
    mov   al,this_sign
#ifdef __BORLANDC__
    mov   esi,[edi]
    mov   [edi.nDim+3],al
#else
    mov   esi,[edi]
    mov   [edi+SIGN],al
#endif
    mov   ebx,abnDigit
  }
IsZero:                       // search nonzero digit
  __asm {
    cmp   dword ptr [ebx*NBYTE+esi-NBYTE],0
    jne   Set
    dec   ebx
  }
Set:
OK:
  __asm {
#ifdef __BORLANDC__
    mov [edi.nDigit],ebx
#else
    mov [edi+NDIGIT],ebx
#endif
  }

RET1:
#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
check(__FILE__,__LINE__) ;
#endif
  return *this ;



//  return isMul(a,a) ;
  }

}
#endif



#if 0
int main(void)
{

    Mint a,f,g,b,c,d ;
    Mint cc ;
    unsigned long m,n ;
    CPUTime cpu1,cpu2 ;
    CPUTime cpu11,cpu22 ;
    CPUTime cpu3 ;
    CPUTime cpu33 ;
    unsigned long seed=time(NULL) ;
    int i,j ;
    FFT fft ;


    a=1 ;
    b=1 ;
    c.isShiftAdd(a,b,1) ;
    component(cout,c) ;
    c.isShiftSub(a,b,1) ;
    component(cout,c) ;
    unsigned long k=0,l ;

#if 1
//  for( k=100 ; k<70000 ; k+=100 ) {
    for( k=160 ; k<2100 ; k++ ) {
        m=irandom(seed) ;
        n=irandom(seed) ;
        m%=20 ;
        n%=50 ;
//      n=m=k ;
//      n=m ;
        cout << endl << "m n " << m << " " << n ;
        cerr << endl << "m n " << m << " " << n ;
//      m=n=k ;
        for( i=0  ; i<2 ; i++ ) {
            a=irandom(seed) ;
            ~a ;
            for( j=1 ; j<m ; j++ ) {
                a[1] ;
                a+=irandom(seed) ;
            }
            b=irandom(seed) ;
            for( j=1 ; j<n ; j++ ) {
                b[1] ;
                b+=irandom(seed) ;
            }
#if 0
        cpu2.Start() ;
        cc.isToom33(a,b) ;
        cpu2.Stop() ;
#endif
            cpu1.Start() ;
            cpu11.Start() ;
            for( j=0;j<100; j++ ) {
            c.isSquare(a) ;
            }
            cpu1.Stop() ;
            cpu11.Stop() ;
//      fft.fft_mul(a,b,c) ;
            cpu3.Start() ;
            cpu33.Start() ;
            for( j=0;j<100; j++ ) {
            d.isSquare2(a) ;
            }
            cpu3.Stop() ;
            cpu33.Stop() ;
//      if( cc!=d ) Error(cc!=d ) ;
#if 0
            if( c!=d ) {
                if( c%a!=0 ) cout << endl << " c%a!=0 " ;
//              if( c%b!=0 ) cout << endl << " c%b!=0 " ;
                if( d%a!=0 ) cout << endl << " d%a!=0 " ;
//              if( d%b!=0 ) cout << endl << " d%b!=0 " ;
                cout << flush ;
                component(cout,c) ;
                component(cout,d) ;
                component(cout,a) ;
//              component(cout,b) ;
                Error(c!=d ) ;
            }
#endif
        }
#endif
        cout << endl << cpu1 ;
        cout << endl << cpu2 ;
        cout << endl << cpu3 ;
        if( cpu3.Time()!=0 ) {
            cout << endl << cpu1.Time()/cpu3.Time() ;
        }

        cerr << endl << cpu1 ;
        cerr << endl << cpu2 ;
        cerr << endl << cpu3 ;
        if( cpu3.Time()!=0 ) {
            cerr << endl << cpu1.Time()/cpu3.Time() ;
        }
        cpu1.Reset() ;
        cpu3.Reset() ;
    }
    cout << endl << cpu11 ;
    cout << endl << cpu2 ;
    cout << endl << cpu33 ;
    if( cpu33.Time()!=0 ) {
        cout << endl << cpu11.Time()/cpu33.Time() ;
    }

cerr << endl <<"final result " ;
    cerr << endl << cpu11 ;
    cerr << endl << cpu2 ;
    cerr << endl << cpu33 ;
    if( cpu33.Time()!=0 ) {
        cerr << endl << cpu11.Time()/cpu33.Time() ;
    }
    return 1 ;
}

#endif
#if 0
#include <stdlib.h>
#include <iostream.h>
#include "Error.hpp" 
#include "IntFuncs.hpp" 
#include <time.h>
int main(void)
{
    Mint a,b,c ;
    Mint s1,s2 ;
    unsigned long s=time(NULL) ;
    unsigned long j,k ;
    unsigned long maxj,maxk ;
    
    a=0xFFFFFFFF ;
//  a=1 ;
    a[1] ;
//  a+=2 ;
//  a+=0xFFFFFFFF ;
    a+=1;
    b.isSquare(a) ;
    c=a*a ;
    component(cout,b) ;
    component(cout,c) ;
    a[1] ;
    a+=0xFFFFFFFF ;
    b.isSquare(a) ;
    c=a*a ;
    component(cout,b) ;
    component(cout,c) ;
    a[1] ;
    a+=1 ;
    b.isSquare(a) ;
    c=a*a ;
    component(cout,b) ;
    component(cout,c) ;
    a[1] ;
    a+=0xFFFFFFFF ;
    b.isSquare(a) ;
    c=a*a ;
    component(cout,b) ;
    component(cout,c) ;
    a[1] ;
    a+=0xFFFFFFFF ;
    b.isSquare(a) ;
    c=a*a ;
    component(cout,b) ;
    component(cout,c) ;
    
    
cout << " k " << k  ;       
    for( k=1 ; k<1000000 ; k++ ) {
        a=irandom(s) ;
cout << " k " << k  ;       
        maxj=irandom(s)%1024 ;
        for( j=0 ; j<maxj ; j++ ) {
            a[1] ;
            a+=irandom(s) ;
        }
        s1=square(a) ;
        s2=a*a ;
        if( s1!=s2 ) Error( why) ;
    }

    return 1 ;

}
#endif