//
//  Copyright (C) by S.Tomabechi 1994-1997
//  All rights reserved.
//
//  add single precision integer to multiple precision number
//
//  Px. _ʒulĉZ.
//
#ifndef __cplusplus
#error Must use C++ for the type Mint
#endif
#ifdef __BORLANDC__
#pragma inline
#endif

#define _ONLY_MINTDECL_
#include "Mint.hpp"

Mint& Mint::isAddUL(const Mint& a,const unsigned long n)
{
#ifdef _MEMMAN_DEBUG
a.check(__FILE__,__LINE__) ;
#endif

  __asm {
    mov ebx,a
#ifdef __BORLANDC__
    cmp dword ptr [ebx.nDim],0      // Is number Positive?
    jge  EXECUTE
    neg byte ptr [ebx.nDim+3]
#else
    cmp dword ptr [ebx+NDIM],0
    jg  EXECUTE
    neg byte ptr [ebx+SIGN]
#endif
  }

  isSubUL(a,n) ;                    // compute as a+n= -((-a)-n)

#ifdef _MEMMAN
_MemMan.check(__FILE__,__LINE__) ;
#endif

  __asm {
    mov ebx,a
    cmp ebx,this
    je  Negate
#ifdef __BORLANDC__
    neg byte ptr [ebx.nDim+3]
#else
    neg byte ptr [ebx+SIGN]
#endif
  }
Negate:
  __asm {
    mov  ebx,this
#ifdef __BORLANDC__
    mov  ecx,[ebx.nDigit]
#else
    mov  ecx,[ebx+NDIGIT]
#endif
    or   ecx,ecx                        // *this==0 ?
    jz   RET0
#ifdef __BORLANDC__
    neg  byte ptr [ebx.nDim+3]
#else
    neg  byte ptr [ebx+SIGN]
#endif
  }

RET0:

#ifdef _MEMMAN
check(__FILE__,__LINE__) ;
#endif
  return *this ;

EXECUTE:
  int shift ;
  int anDigit ;
  int rDigit ;
  __asm {
#ifdef __BORLANDC__
    mov  eax,[ebx.fDigit]
    mov  ecx,[ebx.nDigit]
#else
    mov  eax,[ebx+FDIGIT]
    mov  ecx,[ebx+NDIGIT]
#endif
    mov  shift,eax
    mov  anDigit,ecx
    cmp  ecx,eax
    jg   SET_ECX
    inc  eax
    mov  ecx,eax
  }
SET_ECX:
  __asm {
    mov rDigit,ecx
  }

  if( ndim()<rDigit+1 ) {               // enlarge array
    if( this==&a ) reallocate(rDigit+1) ;
    else allocate(rDigit+1) ;
#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
#endif
  }

  __asm {

    mov   edi,this
    mov   esi,a

    cmp   edi,esi                       // this=&a ?
    je    ADDSELF
#ifdef __BORLANDC__
    mov   ecx,shift                     // copy fractional point
    mov   [edi.fDigit],ecx
    mov   al,[esi.nDim+3]               // copy signature
    mov   [edi.nDim+3],al
    mov   eax,anDigit
    mov   [edi.nDigit],eax      // copy number of digits
    mov   esi,[esi]     // edi=offset of a.digit
    mov   edi,[edi]     // esi=offset of digit
#else
    mov   ecx,shift
    mov   [edi+FDIGIT],ecx
    mov   al,[esi+SIGN]
    mov   [edi+SIGN],al
    mov   eax,anDigit
    mov   [edi+NDIGIT],eax
    mov   esi,[esi]
    mov   edi,[edi]
#endif
    jecxz ADD_N
    cmp   ecx,anDigit  // # significant digits < fractional point  ?
    jbe   Copy
    mov   ecx,anDigit
//    jecxz SET_ZERO
    dec  ecx
    js   SET_ZERO      // in case a=0
  }

Copy:
  __asm {
    mov   eax,[ecx*NBYTE+esi]
    mov   [ecx*NBYTE+edi],eax
    dec   ecx
    jns   Copy
  }
SET_ZERO:
  __asm {
    mov   ecx,shift
    mov   edx,anDigit
    sub   ecx,edx
    jbe   ADD_N
    xor   eax,eax
  }
MORE:
  __asm {
    mov   [edx*NBYTE+edi],eax          // padding 0
    inc   edx
    dec   ecx
    jnz   MORE
  }
                                        // add n
ADD_N:
  __asm {
    mov  ecx,shift
    mov  eax,anDigit
    cmp  ecx,eax
    jb   ADD_NN

    mov   ebx,n
    or    ebx,ebx                       // n=0 ?
    jz    RET1

    mov   [ecx*NBYTE+edi],ebx
    mov   esi,this
    inc   ecx
#ifdef __BORLANDC__
    mov   [esi.nDigit],ecx
#else
    mov   dword ptr [esi+NDIGIT],ecx
#endif
    jmp   RET1
  }
ADD_NN:
  __asm {
    mov   edx,ecx
    sub   ecx,eax
    mov   eax,[edx*NBYTE+esi] // eax=a.digit[shift]
    add   eax,n
    mov   [edx*NBYTE+edi],eax // digit[shift]=eax
    inc   ecx
    jz    Overflow
  }
ADD_CARRY:                              // add carry and copy digits to return value
  __asm {
    inc   edx
    mov   eax,[edx*NBYTE+esi]  // eax <- a.digit[j+shift]
    adc   eax,0                // add carry
    mov   [edx*NBYTE+edi],eax // digit[j+shift] <- eax
    inc   ecx
    js    ADD_CARRY
  }
Overflow:
  __asm {
    jnc   RET1                     // not overflow
  }
Move1:
  __asm {
    inc   edx                           // there is overflow
    mov   dword ptr [edx*NBYTE+edi],1  // set 1 to the most significant digits
    mov   esi,this

#ifdef __BORLANDC__
    inc   dword ptr [esi.nDigit]       // increase # digits
#else
    inc   dword ptr [esi+NDIGIT]
#endif
    jmp   RET1
  }

ADDSELF:                                // in case this==&a
  __asm {                               // faster than in case this!=&a
#ifdef __BORLANDC__
    mov   edi,[edi]
#else
    mov   edi,dword ptr [edi]
#endif
    mov   ecx,shift
    mov   edx,anDigit
    sub   ecx,edx
    jbe   ADD2
  }
MORE2:
  __asm {
    mov   dword ptr [edx*NBYTE+edi],0
    inc   edx
    dec   ecx
    jnz   MORE2
  }


ADD2:

  __asm {
    mov   ecx,shift
    mov   eax,anDigit
    cmp   ecx,eax
    jb    ADD_21

    mov   ebx,n
    or    ebx,ebx
    jz    RET1

    mov   [ecx*NBYTE+edi],ebx

    mov   esi,this
    inc  ecx
#ifdef __BORLANDC__
    mov   [esi.nDigit],ecx
#else
    mov   dword ptr [esi+NDIGIT],ecx
#endif
    jmp   RET1
  }
ADD_21:
  __asm {
    mov   edx,ecx
    sub   ecx,eax

    mov   ebx,n
    add   [edx*NBYTE+edi],ebx
    jnc   RET1                          // no overflow
    inc   ecx
    jecxz Overflow
  }
ADD_LOOP:
  __asm {                               // propagate overflow
    inc   edx
    adc   dword ptr [edx*NBYTE+edi],0  // add carry
    jnc   RET1                           // exit loop if there is no overflow
    inc   ecx
    jnz   ADD_LOOP
    jc    Move1
  }

RET1:
#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
check(__FILE__,__LINE__) ;
#endif
  return *this ;
}

