//
//	Copyright (C) by S.Tomabechi 2000
//	All rights reserved.
//
//	Algorithm 7.4.2 reduction
//	Given tau Im(tau)>0, this algorithm outputs the	 unique tau'
// equivalent to tau under the action of SL_2(Z) and which belongs the
// standard fundamental domain, aw well as the matrix A ( in SZ_2(Z) )
// such that tau' =Atau
#include "Mint.hpp"
#include "Complex.hpp"
#include "Matrix.hpp"

Complex eta(Complex tau);

void quasiPeriod(Complex& w1,Complex& w2,Complex& eta1, Complex& eta2) ;

void reduction(Complex& t,Matrix<Mint>& A)
{
	Mint n ;
	Fint m ;
	Matrix<Mint> B(2,2) ;
	Mint W1,W2 ;
//Initialize
	if( imag(t)<0 ) {
		t=1/t ;
		A(0,0)=A(1,1)=0 ;
		A(0,1)=A(1,0)=1 ;
	}
	else {
		A(0,0)=A(1,1)=1 ;
		A(0,1)=A(1,0)=0 ;
	}
// Reduce real part
	for(;;) {
		n=nearInt(real(t)) ;
		t=t-n ;
		A(0,0)=A(0,0)-n*A(1,0) ;   // translation by n
		A(0,1)=A(0,1)-n*A(1,1) ;
// Finished
		m=norm(t) ;
		if( m>=1 || ndigit(m-1)<2 ) return ;

		t=-conj(t)/m ;
cout << endl << "m " << m ;
if( m==0 ) return ;


		W1=A(0,0) ;
		A(0,0)=-A(1,0) ;   // translation by n
		A(1,0)=W1 ;
		W1=A(0,1)  ;
		A(0,1)=-A(1,1) ;
		A(1,1)=W1 ;
//		A=B ;
	}



}



//
//	Copyright (C) by S.Tomabechi 2001
//	All rights reserved.
//
//	Algorithm 6.3.19 (Aadvanced )
//	Given an oriendted bases (w1,w2) of a complex lattice C and a complex number z.
//	this algorithm computes the value of weierstrass sigma-function sigma(z,L) at z
//
// ܂
#if 0
Complex sigma(Complex& ww1,Complex& ww2,Complex& zz)
{
	Matrix<Mint> A(2,2) ;
	Complex t,At ;		// t =  , At = A
	Complex w1(ww1),w2(ww2),z2 ;
	Complex z(zz) ;
	Mint m,n ;
	Complex eta1,eta2 ;
	Complex s ;
	Complex I(0,1) ;
	Complex y ;
	Complex q8,q ;
	Complex qk ;
	Complex v ;
	Complex vk ;
	Complex S ;
	unsigned long k ;
	Complex e  ;

// Reduce to fundamental domain
	t=w1/w2 ;
	reduction(t,A) ;
	At=(A(0,0)*t+A(0,1))/(A(1,0)*t+A(1,1)) ;   // At is in fundamental domain for SL_2(Z)

	w2=A(1,0)*w1+A(1,1)*w2 ;
	t=At ;
	w1=t*w2 ;
	z2=z/w2 ;

//	Reduce z
	m=nearInt(imag(z2)/imag(t)) ;
	z2-=m*t ;
	n=nearInt(real(z2)) ;
	z2-=n ;
	z=z2*w2 ;

// Compute corrections
//	compute quad-periods eta1,eta2
	quasiPeriod(w1,w2,eta1,eta2) ;
	s=(m*eta1+n*eta2)*(z+(m*w1+n*w2)/2) ;
cout << endl << "m n " << m << " " << n ;
	if( m.isOdd() || n.isOdd() ){
		s+=pi()*I ;
	}

// Initialize
	y=(w2/(2*pi()*I))*exp(s+z2*z*eta2/2) ;
	q8 = exp(2*pi()*I*t/8) ;
//	q=power(q8,8) ;
	((q*=q)*=q)*=q ;
	v=exp(pi()*I*z2) ;

// Compute series

	vk=v ;
	v=v*v ;
	qk=1 ;
	S=0 ;
	for( k=0 ; ; ) {
		S+=(vk-1/vk)*qk ;
//S+=(vk-1/vk)*power(q,k*(k+1)/2) ;
cout << endl << "S " << S ;
		vk*=v ;
		k++ ;
		qk*=power(q,k) ;
		S-=(vk-1/vk)*qk ;
//S-=(vk-1/vk)*power(q,k*(k+1)/2) ;
cout << endl << "S " << S ;
		vk*=v ;
		if( vk==0 ) break ;
		k++ ;
		qk*=power(q,k) ;
		if( qk==0 ) break ;
	}

// terminate
	e=eta(t) ;	// eta function
//cout << endl << "eta " << e ;
	return q8*y*S/(e*e*e) ;

}
#endif



//
//	Copyright (C) by S.Tomabechi 2000
//	All rights reserved.
//
//	Given w1 and w2 generating a lattice L, this algorithm computes the coefficients
//	g2 and g3 f the Weierstrass equation of the elliptic curve C/L
//
//



void g23(Complex& w1,Complex& w2,Complex& g2,Complex& g3)
{
	Complex t,q ;
	Complex I(0,1) ;
	Complex qn ;
	unsigned long n ;
	Matrix<Mint> A(2,2) ;
	Complex w ;
//Initialize
	t=w1/w2 ;
	if( imag(t)<0 ) {
cout << endl << "swap " ;
cout << endl << w1 ;
cout << endl << w2 ;
		Swap(w1,w2) ;
		t=w1/w2 ;
cout << endl << w1 ;
cout << endl << w2 ;
	}

//Reduce
	reduction(t,A) ;
cout <<endl << "post reduction " ;
cout << endl << "t " << t ;
cout << endl <<" A " << A ;
	q=exp(2*pi()*I*t) ;

	g2=0 ;
	g3=0 ;
	qn=1 ;
	Complex c ;
	w=A(1,0)*w2+A(1,1)*w1 ;

	for( n=1 ;; n++ ) {
		qn*=q ;
		if( qn==0 ) break ;
		c=qn/(1-qn) ;
		g2+=power(n,(unsigned long)3)*c ;
		g3+=power(n,(unsigned long)5)*c ;
	}

	g2*=240 ;
	g3*=-504 ;
	g2=g2+1 ;
	g3=g3+1 ;

	g2*=power(2*pi()/w,4) ;
	g2/=12 ;
	g3*=power(2*pi()/w,6) ;
	g3/=216 ;

cout << endl << "last w " ;
cout << endl << w1 ;
cout << endl << w2 ;

}


void pW(Complex& w1,Complex& w2,Complex& z,Complex& p,Complex& dp)
{
	Complex t ;
	Matrix<Mint> A(2,2) ;
	Mint n ;
	Complex I(0,1) ;
	Complex q ;
	Complex u ;
	Complex u2 ;
	Complex qn,qn2 ;
// Initiaalize
	t=w1/w2 ;
	if( imag(t)<0 ) {
		Swap(w1,w2) ;
		t=w1/w2 ;
	}

//Reduce
	reduction(t,A) ;
	t=(A(0,0)*t+A(0,1))/(A(1,0)*t+A(1,1)) ;
	w1=A(1,0)*w2+A(1,1)*w1 ;
	z/=w1 ;
	n=nearInt(imag(z)/imag(t)) ;
	z=z-n*t ;
	z=z-nearInt(real(z)) ;

	if( z==0 ) {

	}
	else {
		q=exp(2*pi()*I*t) ;
		u=exp(2*pi()*I*z/w1) ;
		u2=u*u ;
		qn=1 ;
		for(;;) {
			qn*=q ;
			if( qn==0 ) break  ;
			qn2=qn*qn ;
			p+=qn*(u*(1/(1-2*qn*u+qn2*u2)+1/(qn2-2*qn*u+u2))
					-2/(1-2*qn+qn2)) ;

			dp+=qn*((1+qn)/(1-3*qn*u+3*qn2*u2-qn2*u2*qn*u)
					+(qn+u)/(qn2*qn-3*qn2*u	 + 3*qn*u2-u2*u)) ;
		}
		p+=Fint(1)/12+u/(1-2*u+u2) ;
		dp += (1+u)/(1-3*u+3*u2-u2*u) ;
		dp*=u ;

		p*=power(2*pi()*I/w1,2) ;
		dp*=power(2*pi()*I/w1,3) ;

	}


}


#if 0
int main(void)
{
//	SetPrecision(10) ;
	Complex w1,w2 ;
	Complex g2,g3 ;

	w1=Complex(0,sqrt(Fint(3)))/2;
	w2=Complex(1,0) ;
	g23(w1,w2,g2,g3) ;
	cout << endl << "g2 " << g2 ;
	cout << endl << "g3 " << g3 ;
	cout << endl << "j " << (1728*power(g2,3)/(power(g2,3)-27*g3*g3)) ;


//	pW((w1,w2,p,dp) ;
	return 1 ;
}


#endif


// estimate g2,g3
#if 0
int main(void)
{
//	SetPrecision(10) ;
	Complex w1,w2 ;
	Complex g2,g3 ;


	w1=Complex(0,sqrt(Fint(3)))/2;
	w2=Complex(1,0) ;
	g23(w1,w2,g2,g3) ;
	cout << endl << "g2 " << g2 ;
	cout << endl << "g3 " << g3 ;
	cout << endl << "j " << (1728*power(g2,3)/(power(g2,3)-27*g3*g3)) ;

	w1=Complex(-1,sqrt(Fint(3)))/2;
	w2=Complex(1,0) ;			 // g(,1)=0  =(-1+3)/2
	g23(w1,w2,g2,g3) ;
	cout << endl << "g2 " << g2 ;
	cout << endl << "g3 " << g3 ;
	cout << endl << "j " << (1728*power(g2,3)/(power(g2,3)-27*g3*g3)) ;

	w1=Complex(0,1);
	w2=Complex(1,0) ;	  // g3(O,1)=0
	g23(w1,w2,g2,g3) ;
	cout << endl << "g2 " << g2 ;
	cout << endl << "g3 " << g3 ;
	cout << endl << "j " << (1728*power(g2,3)/(power(g2,3)-27*g3*g3)) ;


//	pW((w1,w2,p,dp) ;
	return 1 ;
}

#endif