//
//  Copyright (C) S.Tomabechi 2001
//  All rights reserved.
//
//  class of primality certificate
//
#ifndef __CERTIFY_HPP
#define __CERTIFY_HPP
#include <vector.h>
#include "Mint.hpp"
using namespace std ;
#define HMAX 10

// fؖ NX
class Certify
{
public:
    static int depth ;
    Mint order ;            // order of elliptic curve
    Mint A ;                // coefficients of elliptic curve (Weierstrass form)
    Mint B ;
    Mint Px ;               // point on elliptic curve
    Mint Py ;
    unsigned long cnum ;
    unsigned long disc ;             // discriminant of complex quadratic field
    Certify* parent ;       // parent of this certificate
    unsigned long hmax ;
    enum ProveMethod { not_prime=0,         // not proved
                        try_div=1,          // proved by trial division
                        Nm1=2,              // proved by N-1
                        Np1=3,              // proved by N+1 (Morrison)
                        KP=4,               // proved by Konyagin and Pomerance
                        MKP=5,              // proved by modified Konyagin and Pomerance
                        APR=6,              // proved by APR-CL
                        BLS=7,              // by BLS
                        ECPP=8,             // by ECPP
                        PRP=9               // probable prime (must be proved by other method)
                        } ;
    enum factorMethod {
            brute=0,                        // gained by brute force
            rho=1,                          // gained by rho-method
            Pollard=2,                      // gained by p-1
            ECM=3,                          // gained by ECM
            MPQS=4,                         // gained by MPQS
            PPSIQS=5,                         // gained by SIQS
            NFS=6 ,                         // gained by NFS
            cofactor=7                      // gained as cofactor
            } ;
    ProveMethod method ;
    factorMethod fmethod ;
    // constructor
    Certify():method(Certify::not_prime),parent(NULL),cnum(1),disc(0),hmax(HMAX) {}
    Certify(const Mint& _N,Certify::ProveMethod m=Certify::not_prime):method(m),
                                parent(NULL),cnum(1),disc(0),hmax(HMAX)
    {
        N=_N ;
    }
    Cerfity(const Certify& C) {
        method=C.method ;
        N=C.N ;
        vm=C.vm ;
        vp=C.vp ;
        parent=C.parent ;
        A=C.A ;
        B=C.B ;
        Px=C.Px ;
        Py=C.Py ;
        order=C.order ;
        disc=C.disc ;
        cnum=C.cnum ;
        hmax=C.hmax ;
//        restbl=C.restbl ;
    }
    // destructor
    ~Certify() {
//        vm.resize(0) ;
//        vp.resize(0) ;
//        ve.resize(0) ;
    }
    // asignment operator
    Certify& operator=(const Certify& c) {
        method=c.method ;
        fmethod=c.fmethod ;
        N=c.N ;
        order=c.order ;
        cnum=c.cnum ;
        disc=c.disc ;
        vm=c.vm ;
        vp=c.vp ;
        ve=c.ve ;
        parent=c.parent ;
        A=c.A ;
        B=c.B ;
        Px=c.Px ;
        Py=c.Py ;
        hmax=c.hmax ;
    }

    Mint  N ;
    vector< Certify > vm ;          // divisors array of N-1
    vector< Certify > vp ;          // divisors array of N+1
    vector< Certify > ve ;          // divisors array of order
    void print(ostream& os) ;
    double factored_Nm1(void) ;
    double factored_Np1(void) ;
    void percent(ostream& os) ;
    bool operator<(const Certify& c) { return N<c.N ; }
    bool operator<=(const Certify& c) { return N<=c.N ; }
    bool operator>(const Certify& c) { return N>c.N ; }
    bool operator>=(const Certify& c) { return N>=c.N ; }
    bool operator==(const Certify& c) { return N==c.N ; }
} ;


ostream& operator<<(ostream& os,const Certify& cert) ;

istream& operator>>(istream& is,Certify& cert) ;

int divisor(Certify& cert,Mint& N, std::vector< Certify >& d) ;
double percent(Mint& N,vector<Certify>& div);
bool PrimeTest(Certify& cert) ;
Mint Rfactor(Mint& N,vector<Certify>& divisor) ;
Mint Rfactor1(Mint& N,vector<Certify>& divisor) ;
double percent1(Mint& N,vector<Certify>& div) ;
bool Pocklington(Certify& C,Mint& B=Mint(1)) ;

#endif