//
//	Copyright (C) by S.Tomabechi 1993,1995
//	All rights reserved.
//
//	base field must be commutative
//
#ifndef __MATRIX_HPP
#define __MATRIX_HPP

#include <vector>
#include <iostream.h>
#include <iomanip.h>

#include "Error.hpp"



template <class T>
class Matrix
{
  friend class Vector ;
private:
  int nDim ;
  int nRow ;
  int nCol ;
  T* pM ;
  bool allocate(const int _ndim) ;
private:
public:
  bool allocate(const int _nrow,const int _ncol) ;
public:
  Matrix() { nDim=nRow=nCol=0 ; pM=(T*)NULL ; }
  Matrix(const int nrow,const int ncol) ;
  Matrix(const Matrix& m) ;
  ~Matrix() ;
  T& operator()(const int r,const int c) ;
  T& operator()(const int r,const int c) const ;
  T* operator[](const int k) ;
  Matrix& operator=(const Matrix& m) ;
  Matrix& operator+=(const Matrix& m) ;
  Matrix  operator+(const Matrix& m) const ;
  Matrix& operator-=(const Matrix& m) ;
  Matrix  operator-(const Matrix& m) const ;
  Matrix& operator*=(const T& s) ;
//	Matrix	operator*(Matrix& m) ;
//	Matrix& operator*=(const Matrix& m) ;
  int nrow(void) const ;
  int ncol(void) const ;

  friend Matrix& operator*=(Matrix& m,const Matrix& n) ;
  friend Matrix operator*(const Matrix& m,const Matrix& n) ;
  friend Matrix operator*(const T& s,const Matrix<T>& m) ;
//	friend Vector operator*(Matrix& m,Vector<T>& v) ;
  friend ::ostream& operator<<(::ostream& os,const Matrix<T>& m) ;
  friend Matrix t(const Matrix& m) ;		// transposed matrix
} ;

template<class T>
bool Matrix<T>::allocate(const int _nrow,const int _ncol)
{
  allocate(_nrow*_ncol) ;
  nRow=_nrow ; nCol=_ncol ;
  return true ;
//	for( int j=nDim-1 ; j>=0 ; j-- ) pM[j]=0 ;	// ͍sȂ
} ;

template<class T>
inline int Matrix<T>::ncol(void) const
{
  return nCol ;
} ;

template<class T>
inline int Matrix<T>::nrow(void) const
{
  return nRow ;
} ;

template<class T>
Matrix<T>& Matrix<T>::operator+=(const Matrix<T>& m)
{
  if( nRow!=m.nRow || nCol!=m.nCol ) Error(not same size) ;
  for( int j=nCol*nRow-1 ; j>=0 ; j-- ) pM[j] += m.pM[j] ;
  return *this ;
} ;

template<class T>
inline Matrix<T> Matrix<T>::operator+(const Matrix<T>& m) const
{
  return Matrix<T>(*this) += m ;
} ;

template<class T>
Matrix<T>& Matrix<T>::operator-=(const Matrix<T>& m)
{
  if( nRow!=m.nRow || nCol!=m.nCol ) Error(not same size) ;
  for( int j=nRow*nCol-1 ; j>=0 ; j-- ) pM[j] -= m.pM[j] ;
  return *this ;
} ;

template<class T>
inline Matrix<T> Matrix<T>::operator-(const Matrix<T>& m) const
{
  return Matrix<T>(*this) -= m ;
} ;


template<class T,class S>
//Matrix<T>& Matrix<T>::operator*=(const Matrix<S>& m)
Matrix<T>& operator*=(Matrix<T>& m,const Matrix<S>& n) 
{
  if( m.nCol!=n.nRow ) Error(size unmatch) ;
  Matrix<T> p(m) ;
  for( int j=0 ; j<n.nRow ; j++ ) {
	for( int k=0 ; k<m.nCol ; k++ ) {
	  m(j,k)=0 ;
	  for( int l=0 ; l<m.nCol ; l++ )
		m(j,k) += p(j,l)*n(l,k) ;
	}
  }
  return m ;
} ;

template<class T>
Matrix<T>& Matrix<T>::operator*=(const T& s)
{
  for( int j=nRow*nCol-1 ; j>=0 ; j-- ) pM[j] *= s ;
  return *this ;
} ;

template<class T>
inline Matrix<T> operator*(const T& s,const Matrix<T>& m)
{
//	Matrix<T> temp(m) ;
//	return temp*=s ;
  return Matrix<T>(m) *= s ;
//	return Matrix<T>(m) *= s ;
//	return (_Mat_temp=m) *= s ;
} ;

template<class T>
inline T* Matrix<T>::operator[](const int r)
{

  return pM+r*nCol ;
} ;

template<class T>
inline T& Matrix<T>::operator()(const int r,const int c)
{
//	if( r*nRow+c>nDim ) Error(index error) ;
//	PRECONDITION(r*nCol+c<=nDim) ;
  return pM[r*nCol+c] ;
} ;

template<class T>
inline T& Matrix<T>::operator()(const int r,const int c) const
{
//	if( r*nRow+c>nDim ) Error(index error) ;
//	PRECONDITION(r*nCol+c<=nDim) ;
  return pM[r*nCol+c] ;
} ;

template<class T>
inline Matrix<T>::~Matrix()
{
  if( pM ) delete [] pM ;
} ;

template<class T>
Matrix<T>::Matrix(const int nrow,const int ncol):pM(NULL)
{
  nDim=nrow*ncol ;
  pM=new T[nDim] ;
  nRow=nrow ;
  nCol=ncol ;
  for( int j=nDim-1 ; j>=0 ; j-- ) pM[j]=0 ;
} ;

template<class T>
inline Matrix<T>::Matrix(const Matrix<T>& m):pM(NULL)
{
  allocate(m.nCol,m.nRow) ;
  *this=m ;
} ;

template<class T>
Matrix<T>& Matrix<T>::operator=(const Matrix<T>& m)
{
  if( this==&m ) return *this ;
  if( nDim<m.nCol*m.nRow ) allocate(m.nCol,m.nRow) ;
  for( int j=nRow*nCol-1 ; j>=0 ; j-- )
	pM[j]=m.pM[j] ;
  nRow=m.nRow ; nCol=m.nCol ;
  return *this ;
} ;

template<class T>
bool Matrix<T>::allocate(const int _ndim)
{
  if( pM ) {
	if( nDim<_ndim ) delete [] pM ;
	else return true ;
  }
  pM=new T[_ndim] ;
  if( !pM ) Error(allocate error) ;
  nDim=_ndim ;
  nRow=_ndim ;
  nCol=1 ;
  for( int j=0 ; j<nDim ; j++ ) pM[j]=0 ;
  return true ;
} ;

//#pragma option -Jgx
template<class T>
inline ::ostream& operator<<(::ostream& os,const Matrix<T>& m)
{
  for( int j=0 ; j<m.nRow ; j++ ) {
	for( int k=0 ; k<m.nCol ; k++ ) {
	  os << m(j,k) << ' '  ;
	}
	os << endl ;
  }
  return os ;
} ;

/**************
template<class T>
Vector<T> operator*(Matrix<T>& m,Vector<T>& v)
{
  Vector<T> mv(v.ndim()) ;
//	PRECONDITION( m.ncol()==v.ndim() ) ;
  for( int j=0 ; j<v.ndim() ; j++ ) {
	mv[j]=0 ;
	for( int k=0 ; k<v.ndim() ; k++ ) mv[j] += m(j,k)*v[k] ;
  }
  return mv ;
} ;
*************/

template<class T>
Matrix<T> t(const Matrix<T>& m)
{
  Matrix<T> tm(m.ncol(),m.nrow());
  int j,k ;
  for( j=tm.nRow-1 ; j>=0 ; j-- ) {
	for( k=tm.nCol-1 ; k>=0 ; k-- )
	  tm(j,k)=m(k,j) ;
  }
  return tm ;
} ;


template<class T>
inline int nrow(Matrix<T>& m)
{
  return m.nrow() ;
} ;

template<class T>
inline int ncol(Matrix<T>& m)
{
  return m.ncol() ;
} ;
//#pragma option -Jgx.

template<class T,class S>
Matrix<T> operator*(const Matrix<T>& m,const Matrix<S>& n)
{
  if( m.ncol()!=n.nrow() ) Error(size unmatch) ;
  Matrix<T> mn(m.nrow(),n.ncol()) ;
  int j,k,l ;
  for( j=m.nrow()-1 ; j>=0 ; j-- ) {
	for( k=n.ncol()-1 ; k>=0 ; k-- ) {
	  mn(j,k)=0 ;
	  for( l=m.ncol()-1 ; l>=0 ; l-- )
		mn(j,k) += m(j,l)*n(l,k) ;
	}
  }
  return mn ;
} ;


//using namespace std ;
#if 0
template<class T>
ostream& operator<<(ostream& os,const std::vector<T>& a)
{
  for( int j=0 ; j<a.size() ; j++ ) os << a[j] << " "  ;
  return os ;
} ;
#endif
template<class T>
std::vector<T> operator*(Matrix<T>& m,std::vector<T>& v)
{
	std::vector<T> mv ;
	mv.resize(m.nrow()) ;
	for( int j=0 ; j<m.nrow() ; j++ ) {
		mv[j]=0 ;
		for( int k=0 ; k<m.ncol() ; k++ ) {
			mv[j]+=m(j,k)*v[k] ;
		}
	}
	return mv ;
} ;

#endif
